﻿/********************************************************
 *  ██████╗  ██████╗████████╗██╗
 * ██╔════╝ ██╔════╝╚══██╔══╝██║
 * ██║  ███╗██║        ██║   ██║
 * ██║   ██║██║        ██║   ██║
 * ╚██████╔╝╚██████╗   ██║   ███████╗
 *  ╚═════╝  ╚═════╝   ╚═╝   ╚══════╝
 * Geophysical Computational Tools & Library (GCTL)
 *
 * Copyright (c) 2023  Yi Zhang (yizhang-geo@zju.edu.cn)
 *
 * GCTL is distributed under a dual licensing scheme. You can redistribute 
 * it and/or modify it under the terms of the GNU Lesser General Public 
 * License as published by the Free Software Foundation, either version 2 
 * of the License, or (at your option) any later version. You should have 
 * received a copy of the GNU Lesser General Public License along with this 
 * program. If not, see <http://www.gnu.org/licenses/>.
 * 
 * If the terms and conditions of the LGPL v.2. would prevent you from using 
 * the GCTL, please consider the option to obtain a commercial license for a 
 * fee. These licenses are offered by the GCTL's original author. As a rule, 
 * licenses are provided "as-is", unlimited in time for a one time fee. Please 
 * send corresponding requests to: yizhang-geo@zju.edu.cn. Please do not forget 
 * to include some description of your company and the realm of its activities. 
 * Also add information on how to contact you by electronic and paper mail.
 ******************************************************/

#ifndef _GCTL_REGULAR_GRID_H
#define _GCTL_REGULAR_GRID_H

#include "gctl/graphic.h"
#include "mesh.h"

#ifdef GCTL_MESH_EXPRTK
#include "exprtk.hpp"
#endif // GCTL_MESH_EXPRTK

namespace gctl
{
	/**
	 * @brief 规则网格对象
	 * 
	 */
	class regular_grid : public base_mesh
	{
	public:

		/**
		 * 以下为mesh类型的虚函数实现
		 */

		void init(std::string in_name, std::string in_info, int xnum, int ynum, 
			double xmin, double ymin, double dx, double dy);

		void show_mesh_dimension(std::ostream &os) const;

		void load_binary(std::string filename);
		void save_binary(std::string filename);

		/**
		 * 以下为regular_grid的专有函数
		 */

		regular_grid();
		regular_grid(std::string in_name, std::string in_info, int xnum, int ynum, 
			double xmin, double ymin, double dx, double dy);
		virtual ~regular_grid();

		void clear_regular_grid();
		int view(std::string datname);
		void plot(std::string datname);

		int get_xdim() const;
		int get_ydim() const;
		double get_xmin() const;
		double get_ymin() const;
		double get_dx() const;
		double get_dy() const;

#ifdef GCTL_NETCDF

		void load_netcdf_grid(std::string filename, mesh_data_type_e d_type, 
			std::string xname = "x", std::string yname = "y", std::string zname = "null");
		void save_netcdf_grid(std::string filename, mesh_data_type_e d_type, 
			std::string xname = "x", std::string yname = "y");
		void save_netcdf_grid(std::string filename, std::string datname, 
			std::string xname = "x", std::string yname = "y");

#endif // GCTL_NETCDF

		void load_surfer_grid(std::string filename, std::string datname, mesh_data_type_e d_type, surfer_file_type_e grid_type = Surfer7Grid);
		void save_surfer_grid(std::string filename, std::string datname, surfer_file_type_e grid_type = Surfer7Grid);

		void save_gmsh(std::string filename, index_packed_e packed = Packed);
		void save_gmsh(std::string filename, mesh_data_type_e d_type, output_type_e out_mode, index_packed_e packed = Packed);
		void save_gmsh(std::string filename, std::string datname, output_type_e out_mode, index_packed_e packed = Packed);

		void save_text(std::string filename, const array<std::string> &datname);

		void load_data_cloud(const array<point2dc> &in_posi, const array<double> &in_val, double search_xlen, 
			double search_ylen, double search_deg, std::string datname, mesh_data_type_e d_type = NodeData);

		void extract_points(std::string datname, const array<point2dc> &in_posi, array<double> &out_val);
		void extract_profile(std::string datname, const point2dc &start_p, const point2dc &end_p, int size_p, 
			array<point2dc> &out_posi, array<double> &out_val);

		void gradient(std::string datname, std::string gradname, gradient_type_e d_type, int order = 1);
		void sum(std::string newname, std::string datname, std::string datname2);
		void diff(std::string newname, std::string datname, std::string datname2);
		void boolean(std::string newname, std::string datname, std::string maskname, bool reverse = false);

#ifdef GCTL_MESH_EXPRTK
		void function(std::string expression_str, std::string newname, std::string x_str = "x", std::string y_str = "y", std::string f_str = "f");
		void calculator(std::string expression_str, const array<std::string> &var_list, const array<std::string> &data_list);
#endif // GCTL_MESH_EXPRTK

#ifdef GCTL_MESH_WAVELIB
		void wavelet(std::string datname, std::string wavename, int order, bool summary = true);
#endif // GCTL_MESH_WAVELIB

	protected:
		int rg_xnum, rg_ynum;
		double rg_xmin, rg_ymin;
		double rg_dx, rg_dy;

#ifdef GCTL_GRAPHIC_MATHGL
		mathgl_dens plt;
#endif // GCTL_GRAPHIC_MATHGL

#ifdef GCTL_GRAPHIC_GMT
		gmt_JX_single pic;
#endif // GCTL_GRAPHIC_GMT

		array<vertex2dc> nodes;
		array<rectangle2d> elements;
	};
}

#endif //_GCTL_REGULAR_GRID_H
